use anyhow::Result;
use std::{
    ops::{Deref, DerefMut},
    sync::{Arc, Mutex, MutexGuard, Weak},
};

#[derive(Default)]
pub struct EngineObjectData<T> {
    data: Arc<Mutex<T>>,
}

impl<T> EngineObjectData<T> {
    pub fn new(t: T) -> Self {
        Self {
            data: Arc::new(Mutex::new(t)),
        }
    }

    pub fn access(&self) -> EngineObjectAccess<'_, T> {
        EngineObjectAccess {
            data: self.data.lock().unwrap(),
        }
    }

    pub fn copy(&self) -> EngineObjectDataHandle<T> {
        EngineObjectDataHandle {
            data: Arc::downgrade(&self.data),
        }
    }
}

impl<T> From<T> for EngineObjectData<T> {
    fn from(t: T) -> Self {
        Self {
            data: Arc::new(Mutex::new(t)),
        }
    }
}

pub struct EngineObjectDataHandle<T> {
    data: Weak<Mutex<T>>,
}

impl<T> EngineObjectDataHandle<T>
where
    T: Send + Sync + 'static,
{
    pub fn as_data(&self) -> EngineObjectData<T> {
        EngineObjectData {
            data: self.data.upgrade().unwrap(),
        }
    }

    pub(crate) fn copy(&self) -> Self {
        Self {
            data: self.data.clone(),
        }
    }

    pub fn access<F>(&self, mut f: F) -> Result<()>
    where
        F: FnMut(EngineObjectAccess<'_, T>) -> Result<()> + Send + Sync + 'static,
    {
        let data = self.as_data();

        f(data.access())
    }

    pub fn access_once<'a, F>(&self, f: F) -> Result<()>
    where
        F: FnOnce(EngineObjectAccess<'_, T>) -> Result<()> + Send + Sync + 'a,
    {
        let data = self.as_data();

        f(data.access())
    }

    pub fn access_callback_mut<F>(
        &self,
        mut f: F,
    ) -> impl FnMut() -> Result<()> + Send + Sync + 'static
    where
        F: FnMut(EngineObjectAccess<'_, T>) -> Result<()> + Send + Sync + 'static,
    {
        let handle = self.copy();

        move || {
            let data = handle.as_data();
            let access = data.access();

            f(access)
        }
    }

    pub fn access_callback<F>(&self, f: F) -> impl Fn() -> Result<()> + Send + Sync + 'static
    where
        F: Fn(EngineObjectAccess<'_, T>) -> Result<()> + Send + Sync + 'static,
    {
        let handle = self.copy();

        move || {
            let data = handle.as_data();
            let access = data.access();

            f(access)
        }
    }
}

impl<T> Clone for EngineObjectDataHandle<T> {
    fn clone(&self) -> Self {
        Self {
            data: self.data.clone(),
        }
    }
}

pub struct EngineObjectAccess<'a, T> {
    data: MutexGuard<'a, T>,
}

impl<'a, T> Deref for EngineObjectAccess<'a, T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.data
    }
}

impl<'a, T> DerefMut for EngineObjectAccess<'a, T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.data
    }
}
