use anyhow::Result;
use cgmath::Vector3;
use engine::prelude::*;

use std::sync::Arc;

use crate::Game;

use super::health_bar_base::HealthBarBase;

pub struct UserHealthBar {
    gui: Arc<GuiBuilder>,

    name_grid: Arc<Grid>,
    bar_grid: Arc<Grid>,

    hp_bar: Arc<ProgressBar>,
    mana_bar: Arc<ProgressBar>,

    base: HealthBarBase,
}

impl UserHealthBar {
    pub fn new(game: &Game, user_name: &str, world_offset: Vector3<f32>) -> Result<Self> {
        let gui = GuiBuilder::new(
            game.engine().gui_handler(),
            &game.build_data_path("gui/xml/ingame/user_healthbar.xml"),
        )?;

        let hp_bar: Arc<ProgressBar> = gui.element("health bar")?;
        hp_bar.set_progress(1.0)?;

        let mana_bar: Arc<ProgressBar> = gui.element("mana bar")?;
        mana_bar.set_progress(1.0)?;

        let name_grid: Arc<Grid> = gui.element("name grid")?;

        // this is super important to prevent weird behavior
        name_grid.disallow_position_scale()?;

        let bar_grid: Arc<Grid> = gui.element("bar grid")?;

        // this is super important to prevent weird behavior
        bar_grid.disallow_position_scale()?;

        let user_name_label: Arc<Label> = gui.element("user name")?;
        user_name_label.set_text(&user_name)?;

        Ok(UserHealthBar {
            gui,

            name_grid,
            bar_grid,

            hp_bar,
            mana_bar,

            base: HealthBarBase {
                center: (0, 0),

                should_enable: false,
                is_enabled: false,

                world_offset,
            },
        })
    }

    pub fn update(
        &mut self,
        hp_progress: f32,
        mana_progress: f32,
        (x, y): (i32, i32),
    ) -> Result<()> {
        if self.base.should_enable && !self.base.is_enabled {
            self.base.is_enabled = true;
            self.gui.enable()?;
        }

        self.hp_bar.set_progress(hp_progress)?;
        self.mana_bar.set_progress(mana_progress)?;

        if self.base.center != (x, y) {
            self.base.center = (x, y);

            let (name_width, name_height) = self.name_grid.extents();
            let (bar_width, _) = self.bar_grid.extents();

            self.name_grid
                .set_position(x - (name_width as i32 / 2), y - (name_height as i32 / 2))?;

            self.bar_grid
                .set_position(x - (bar_width as i32 / 2), y + (name_height as i32 / 2))?;
        }

        Ok(())
    }

    pub fn world_offset(&self) -> Vector3<f32> {
        self.base.world_offset
    }
}

impl EntityComponent for UserHealthBar {
    fn enable(&mut self, __scene: &mut Scene) -> Result<()> {
        self.base.should_enable = true;

        Ok(())
    }

    fn disable(&mut self, _scene: &mut Scene) -> Result<()> {
        self.base.should_enable = false;
        self.base.is_enabled = false;

        self.gui.disable()?;

        Ok(())
    }

    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for UserHealthBar {
    fn debug_name() -> &'static str {
        "UserHealthBar"
    }
}
